package com.noorija.secret.guard.service

import android.app.Notification
import android.app.NotificationChannel
import android.app.NotificationManager
import android.app.PendingIntent
import android.app.Service
import android.content.Intent
import android.os.Binder
import android.os.Build
import android.os.Bundle
import android.os.IBinder
import android.os.ResultReceiver
import android.util.Log

import androidx.annotation.Nullable
import com.google.gson.GsonBuilder
import com.noorija.secret.guard.R
import com.noorija.secret.guard.manager.FindCallRtcManager
import com.noorija.secret.guard.manager.FindCallStompManger
import com.noorija.secret.guard.model.StompPayload
import com.noorija.secret.guard.net.OnStompMessageListener
import com.noorija.secret.guard.view.MainActivity
import com.noorija.secret.guard.view.RcvMsgActivity
import timber.log.Timber

class FindCallService: Service(), OnStompMessageListener {
    companion object {
        private const val CHANNEL_ID: String = "ForegroundServiceChannel"

        const val MSG_STOMP_MESSAGE = 3
        const val MSG_STOMP_CALL = 4
    }

    // 서비스 바인딩을 위한 Binder 클래스
    private val localBinder by lazy { LocalBinder() }

    private val rtcManager by lazy {
        FindCallRtcManager.createFindCallRtcManager(this)
    }

    private val stompManager by lazy {
        FindCallStompManger.createFindCallStompManger(this)
    }

    @Nullable
    override fun onBind(intent: Intent?): IBinder? {
        Log.i("FindCallService ", "onBind")
        return localBinder
//        return null
    }

    // 오버라이딩, 함수 재정의
    // 서비스가 생성될 때 실행할 작업을 정의합니다
    override fun onCreate() {
        // 부모의 onDestroy 함수 호출
        super.onCreate()
        Log.i("FindCallService ", "onCreate")
        createNotificationChannel()
    }

    private fun createNotificationChannel() {
        if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.O) {
            //            첫 번째 인수는 채널 ID를 나타내며, 두 번째 인수는 채널 이름을 지정합니다. 세 번째 인수는 채널의 중요도를 지정합니다
            val serviceChannel = NotificationChannel(
                CHANNEL_ID,
                "Foreground Service Channel",
                NotificationManager.IMPORTANCE_DEFAULT
            )

            val manager = checkNotNull(
                getSystemService(
                    NotificationManager::class.java
                )
            )
            //새로운 알림 채널을 등록
            manager.createNotificationChannel(serviceChannel)
        }
    }

    // 오버라이딩, 함수 재정의
    // 서비스가 시작될 때 실행할 작업을 정의합니다.
    // 메인 작업 처리?
    override fun onStartCommand(intent: Intent, flags: Int, startId: Int): Int {
        Log.i("FindCallService ", "onStartCommand")
        val notificationIntent = Intent(this, MainActivity::class.java)
        val pendingIntent =
            PendingIntent.getActivity(
                this, 0, notificationIntent,
                PendingIntent.FLAG_IMMUTABLE
            )

        //        setContentTitle(), setContentText(), setSmallIcon() 메서드를 사용하여 알림의 제목, 내용 및 작은 아이콘을 설정합니다.
//        setContentIntent() 메서드를 사용하여 알림을 탭했을 때 실행될 인텐트를 설정합니다.
        var notification: Notification? =
            null
        if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.O) {
            notification = Notification.Builder(applicationContext, CHANNEL_ID)
                .setContentTitle("제목1") //getText(R.string.notification_title)
                .setContentText("내용")
                .setSmallIcon(R.drawable.ico_logo)
                .setContentIntent(pendingIntent) //                    .setTicker("Ticker")
                //                    .setAutoCancel(true)
                .build()

            startForeground(1, notification)
            Log.i("FindCallService ", "노티 시작")
        }

        Log.i("FindCallService : ", notification.toString())

        val bundle = Bundle()
        //ResultReceiver를 가져온다.
        //Parcelable를 상속받아서 구현하는 ResultReceiver
        //ResultReceiver : 백그라운드 서비스에서 작업을 수행하고, 결과를 메인 스레드에 반환하는 데 사용
        val receiver: ResultReceiver? = intent.getParcelableExtra(MainActivity.CMD_FIND_CALL_RECEIVER) // (1)
        bundle.putString("msg", "Succeed !")

        // send() 메서드를 호출하여 결과를 반환하면 onReceiveResult() 메서드가 자동으로 호출
        //결과 코드(resultCode)와 결과 데이터(bundle)를 인수로 받습니다.
        receiver?.send(1, bundle) // (2)
//        intent.getParcelableExtra<ResultReceiver?>("RECEIVER")?.send(1, bundle) // (2)

//        rtcManager.initModule()
        intent.getStringExtra(MainActivity.CMD_ROOM_ID_LIST)?.let { stompManager.init(it) }

        stompManager.addListener(this)

        // 부모 onStartCommand 호출, 반환.
        return super.onStartCommand(intent, flags, startId)
        // onStartCommand 함수의 끝.
    }

    // 오버라이딩, 함수 재정의
    // 서비스가 종료될 때 실행할 작업을 정의합니다.
    override fun onDestroy() {
        // 부모의 onDestroy 함수 호출
        super.onDestroy()
        // MyService onDestroy 출력
        Log.i("FindCallService", "onDestroy")
        // onDestroy 함수 끝.
    }

    internal inner class LocalBinder : Binder() {
        fun getService(): FindCallService = this@FindCallService
    }

    fun bindMessage(bindType: Int, message: String, strPayload: String) {
        if(bindType == 0) {
            stompManager.sendStompMessage(message, strPayload)
        } else if(bindType == 1) {

        }
    }

    override fun onMessageReceived(nickNm: String, message: StompPayload) {
        Timber.d("FindCallService onMessageReceived >>>>>>>>>>>>>>>>>>>>>>>>>>>>>> $message")
        when(message.mType) {
            "C" -> { }  // 통화 요청시 key 전달건
            "M" -> { onRcvMsg(nickNm, message) }  // 문자 수신 받았을 경우
            "BR" -> { }
            "BC" -> { } // 동작을 안함 - 무시
            "BJ" -> { } // webRTC 접속 취소 pc ==> 전화거절 화면으로 이동한다ㅏ.
            "BT" -> { } // 전화를 문자로 답변
            "O" -> { onRcvMsg(nickNm, message) }  //상대 접속 종료로 인한 통신 종료 처리
        }
    }

    private fun onRcvMsg(nickNm: String, message : StompPayload) {
        Timber.d("onRcvMsg >>> $message")
        val gson = GsonBuilder()
            .disableHtmlEscaping()
            .serializeNulls()
            .create()
        PendingIntent.getActivity(
            this@FindCallService,
            220,
            Intent(this@FindCallService, RcvMsgActivity::class.java).apply {
                this.flags = Intent.FLAG_ACTIVITY_SINGLE_TOP
                putExtra("StompNickName", nickNm)
                putExtra("StompPayload", gson.toJson(message, StompPayload::class.java))
            },
            PendingIntent.FLAG_UPDATE_CURRENT
        ).send()
    }

}