package com.noorija.secret.guard.viewmodel

import android.annotation.SuppressLint
import android.app.Application
import android.content.Context
import android.content.pm.PackageManager
import android.view.Gravity
import android.view.LayoutInflater
import android.view.View
import android.widget.Toast
import androidx.appcompat.app.AlertDialog
import androidx.core.content.ContextCompat
import androidx.lifecycle.AndroidViewModel
import androidx.lifecycle.MutableLiveData
import com.gun0912.tedpermission.PermissionListener
import com.gun0912.tedpermission.normal.TedPermission
import com.noorija.secret.guard.util.SystemUtil
import io.reactivex.Single
import io.reactivex.SingleObserver
import io.reactivex.android.schedulers.AndroidSchedulers
import io.reactivex.disposables.CompositeDisposable
import io.reactivex.disposables.Disposable
import io.reactivex.observers.DisposableSingleObserver
import io.reactivex.schedulers.Schedulers
import com.noorija.secret.guard.R
import com.noorija.secret.guard.databinding.DlgCustomPopupBinding
import com.noorija.secret.guard.rest.ExceptionDTO
import com.noorija.secret.guard.rest.RetrofitClient
import com.noorija.secret.guard.rest.RetrofitInterface
import com.noorija.secret.guard.text.Command
import com.noorija.secret.guard.text.DialogResultType
import com.noorija.secret.guard.text.DialogType
import com.noorija.secret.guard.util.PreferencesUtil
import retrofit2.HttpException

@SuppressLint("StaticFieldLeak")
abstract class BaseViewModel<D> protected constructor(application: Application): AndroidViewModel(application) {
    private val disposable: CompositeDisposable = CompositeDisposable()
    var isLoading: MutableLiveData<Boolean> = MutableLiveData()
    val context: Context = getApplication<Application>().applicationContext
    private var toast: Toast? = null

    val restClient: RetrofitInterface = RetrofitClient.createInterface()

    override fun onCleared() {
        disposable.clear()
        super.onCleared()
    }

    fun addDisposable(single: Single<*>, observer: DisposableSingleObserver<*>, isDialog: Boolean = false) {
        if(SystemUtil.isConnected(context)) {
            isLoading.value = true

            disposable.add(single.subscribeOn(Schedulers.io())
                .observeOn(AndroidSchedulers.mainThread()).subscribeWith(observer as SingleObserver<Any>) as Disposable
            )
        } else {
            showToast(R.string.error_network_state)
        }
    }

    val dataObserver: DisposableSingleObserver<D>
        get() = object : DisposableSingleObserver<D>() {

            override fun onSuccess(t: D & Any) {
                try {
                    onRetrieveDataSuccess(t)
                } catch (ex: Exception) { }
                isLoading.value = false
            }

            override fun onError(e: Throwable) {
                try {
                    val error: HttpException = e as HttpException
                    val errorBody = error.response()?.errorBody()?.string()
                    val resError = SystemUtil.jsonToObj(errorBody, ExceptionDTO::class.java)
                    onRetrieveDataError(resError)
                } catch (ex: Exception) {
                    println(ex.printStackTrace())
                }
                isLoading.value = false
            }
        }

    protected fun isBlankAppKey(): Boolean {
        return PreferencesUtil.getString(Command.SECRET_GUARD_PREF_NAME_APP_KEY).isEmpty()
    }

    protected fun getToken(): String {
        val token = PreferencesUtil.getString(Command.SECRET_GUARD_PREF_NAME_APP_KEY)
        return "Bearer $token"
    }

    protected open fun onRetrieveDataSuccess(data: D) { }
    protected open fun onRetrieveDataError(resError: ExceptionDTO?) { }

    protected fun requestPermissions(permission: Any, callback : (isPermission: Boolean) -> Unit) {
        val permissions = TedPermission.create()
            .setPermissionListener(object : PermissionListener {
                override fun onPermissionGranted() {
                    callback(true)
                }
                override fun onPermissionDenied(deniedPermissions: List<String>) {
                    showToast(R.string.msg_permission_request)
                    callback(false)
                }
            })
            .setRationaleMessage("")
            .setDeniedMessage(R.string.msg_permission_denied)
        val permissionList = ArrayList<String>()
        if(permission is String) {
            if(ContextCompat.checkSelfPermission(context, permission) != PackageManager.PERMISSION_GRANTED) {
                permissionList.add(permission)
            }
        } else {
            for(item: String in permission as Array<String>) {
                if(ContextCompat.checkSelfPermission(context, item) != PackageManager.PERMISSION_GRANTED) {
                    permissionList.add(item)
                }
            }
        }
        if(permissionList.size > 0) {
            permissions.setPermissions(*permissionList.toTypedArray())
            permissions.check()
        } else {
            callback(true)
        }

//        when(permission) {
//            is String   -> permissions.setPermissions(permission)
//            else        -> permissions.setPermissions(*checkPermission(permission as Array<String>))
//        }
//        permissions.check()
    }

    private fun checkPermission(permissions: Array<String>): Array<String> {
        val permissionList = ArrayList<String>()
        for(permission: String in permissions) {
            if(ContextCompat.checkSelfPermission(context, permission) != PackageManager.PERMISSION_GRANTED) {
                permissionList.add(permission)
            }
        }
        return permissionList.toTypedArray()
    }

    protected fun isPermission(permission : String): Boolean {
        if(ContextCompat.checkSelfPermission(context, permission) != PackageManager.PERMISSION_GRANTED) {
            return true
        }
        return false
    }

    protected fun showToast(message: String?, gravity: Int = Gravity.BOTTOM) {
        toast = Toast.makeText(context, message, Toast.LENGTH_SHORT)
        toast?.setGravity(gravity, 0, 0)
        toast?.show()
    }

    protected fun showToast(resId: Int, gravity: Int = Gravity.BOTTOM) {
        toast = Toast.makeText(context, resId, Toast.LENGTH_SHORT)
        toast?.setGravity(gravity, 0, 0)
        toast?.show()
    }

    open fun showCustomDialog(
        id: Int,
        message: String,
        dialogType: DialogType = DialogType.ALERT,
        title: String = "",
        btnLeftText: String = "",
        btnRightText: String = "",
        data: Any
    ) {
        val dialogView = DlgCustomPopupBinding.inflate(LayoutInflater.from(context))
        val builder = AlertDialog.Builder(context).setView(dialogView.root)
        val alertDialog = builder.create()
        alertDialog.show()

        if (title.isEmpty())
            dialogView.title.visibility = View.GONE
        else
            dialogView.title.text = title
        dialogView.message.text = message
        if (dialogType == DialogType.ALERT) {
            dialogView.btnCancel.visibility = View.GONE
        } else {
            if (btnLeftText.isNotEmpty()) dialogView.btnCancel.text = btnLeftText
        }
        if (btnRightText.isNotEmpty()) dialogView.btnConfirm.text = btnRightText
        dialogView.btnConfirm.setOnClickListener {
            alertDialog.dismiss()
            onDialogResult(id, data, DialogResultType.CONFIRM)
        }
        dialogView.btnCancel.setOnClickListener {
            alertDialog.dismiss()
            onDialogResult(id, DialogResultType.CANCEL)
        }
    }

    protected open fun onDialogResult(id: Int, result: DialogResultType) { }

    protected open fun onDialogResult(id: Int, data: Any, result: DialogResultType) { }

}